/*
 * idna.c - IDNA conversion DLL for mIRC
 * Copyright (c) 2004, D.C. van Moolenbroek, XISE <dc@xise.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Contains code from:
 *
 * > punycode.c from RFC 3492
 * > http://www.nicemice.net/idn/
 * > Adam M. Costello
 * > http://www.nicemice.net/amc/
 * >
 * > This is ANSI C code (C89) implementing Punycode (RFC 3492).
 *
 */

#include <windows.h>

#define basic(cp) ((DWORD)(cp) < 0x80)
#define delim(cp) ((cp) == delimiter)

enum { base = 36, tmin = 1, tmax = 26, skew = 38, damp = 700,
       initial_bias = 72, initial_n = 0x80, delimiter = 0x2D };

static const DWORD maxint = -1;

static DWORD decode_digit(DWORD cp) {
  return  cp - 48 < 10 ? cp - 22 :  cp - 65 < 26 ? cp - 65 :
          cp - 97 < 26 ? cp - 97 :  base;
}

static char encode_digit(DWORD d, int flag) {
  return d + 22 + 75 * (d < 26) - ((flag != 0) << 5);
}

static DWORD adapt(DWORD delta, DWORD numpoints, int firsttime) {
  DWORD k;

  delta = firsttime ? delta / damp : delta >> 1;
  delta += delta / numpoints;

  for (k = 0;  delta > ((base - tmin) * tmax) / 2;  k += base) {
    delta /= base - tmin;
  }

  return k + (base - tmin + 1) * delta / (delta + skew);
}

static int punycode_encode(DWORD input_length, const wchar_t input[],
 DWORD *output_length, TCHAR output[]) {
  DWORD n, delta, h, b, out, max_out, bias, j, m, q, k, t;

  n = initial_n;
  delta = out = 0;
  max_out = *output_length;
  bias = initial_bias;

  for (j = 0;  j < input_length;  ++j) {
    if (basic(input[j])) {
      if (max_out - out < 2) return -1;
      output[out++] = input[j];
    }
  }

  h = b = out;

  if (b > 0) output[out++] = delimiter;

  while (h < input_length) {
    for (m = maxint, j = 0;  j < input_length;  ++j) {
      if ((DWORD)input[j] >= n && (DWORD)input[j] < m) m = input[j];
    }

    if (m - n > (maxint - delta) / (h + 1)) return -1;
    delta += (m - n) * (h + 1);
    n = m;

    for (j = 0;  j < input_length;  ++j) {
      if ((DWORD)input[j] < n) {
        if (++delta == 0) return -1;
      }

      if ((DWORD)input[j] == n) {
        for (q = delta, k = base;  ;  k += base) {
          if (out >= max_out) return -1;
          t = k <= bias ? tmin : k >= bias + tmax ? tmax : k - bias;
          if (q < t) break;
          output[out++] = encode_digit(t + (q - t) % (base - t), 0);
          q = (q - t) / (base - t);
        }

        output[out++] = encode_digit(q, 0);
        bias = adapt(delta, h + 1, h == b);
        delta = 0;
        ++h;
      }
    }

    ++delta, ++n;
  }

  *output_length = out;
  return 0;
}

static int punycode_decode(DWORD input_length, const TCHAR input[],
 DWORD *output_length, wchar_t output[]) {
  DWORD n, out, i, max_out, bias, b, j, in, oldi, w, k, digit, t;

  n = initial_n;
  out = i = 0;
  max_out = *output_length;
  bias = initial_bias;

  for (b = j = 0;  j < input_length;  ++j) if (delim(input[j])) b = j;
  if (b > max_out) return -1;

  for (j = 0;  j < b;  ++j) {
    if (!basic(input[j])) return -1;
    output[out++] = input[j];
  }

  for (in = b > 0 ? b + 1 : 0;  in < input_length;  ++out) {
    for (oldi = i, w = 1, k = base;  ;  k += base) {
      if (in >= input_length) return -1;
      digit = decode_digit(input[in++]);
      if (digit >= base) return -1;
      if (digit > (maxint - i) / w) return -1;
      i += digit * w;
      t = k <= bias ? tmin : k >= bias + tmax ? tmax : k - bias;
      if (digit < t) break;
      if (w > maxint / (base - t)) return -1;
      w *= (base - t);
    }

    bias = adapt(i - oldi, out + 1, oldi == 0);

    if (i / (out + 1) > maxint - n) return -1;
    n += i / (out + 1);
    i %= (out + 1);

    if (out >= max_out) return -1;

    MoveMemory(output + i + 1, output + i, (out - i) * sizeof *output);
    output[i++] = n;
  }

  *output_length = out;
  return 0;
}

int __export __stdcall encode(HWND mWnd, HWND aWnd, TCHAR *data,
 TCHAR *parms, BOOL show, BOOL nopause) {
  wchar_t ebuf[256];
  TCHAR obuf[1023];
  DWORD in, out, end, len, elen, flag;

  for (in = out = 0; data[in]; in += end + 1) {
    for (end = flag = 0; data[in+end] && data[in+end] != '.'; end++)
     if (!basic(data[in+end])) flag = 1;

    if (flag) {
#ifndef UNICODE
      elen = MultiByteToWideChar(CP_ACP, 0, data + in, end, ebuf,
       sizeof(ebuf) >> 1);
#else
      elen = sizeof(ebuf) >> 1;
      if (elen > end) elen = end;
      CopyMemory(ebuf, data + in, elen << 1);
#endif

      if (!elen || out + 4 >= sizeof(obuf) / sizeof(TCHAR)) {
        out = 0;
        break;
      }

      lstrcpy(obuf + out, TEXT("xn--"));
      out += 4;
      len = sizeof(obuf) / sizeof(TCHAR) - out - 1;

      if (punycode_encode(elen, ebuf, &len, obuf + out)) {
        out = 0;
        break;
      }

      out += len;
    }
    else {
      if (out + end + 1 >= sizeof(obuf) / sizeof(TCHAR)) {
        out = 0;
        break;
      }
      CopyMemory(obuf + out, data + in, end * sizeof(TCHAR));
      out += end;
    }

    if (!data[in+end]) break;
    obuf[out++] = data[in+end];
  }

  CopyMemory(data, obuf, out * sizeof(TCHAR));
  data[out] = 0;
  return 3;
}

static int isxnhdr(TCHAR *ptr) {
  if ((ptr[0] == 'X' || ptr[0] == 'x') &&
      (ptr[1] == 'N' || ptr[1] == 'n') &&
      ptr[2] == '-' && ptr[3] == '-') return 1;
  return 0;
}

int __export __stdcall decode(HWND mWnd, HWND aWnd, TCHAR *data,
 TCHAR *parms, BOOL show, BOOL nopause) {
  wchar_t dbuf[256];
  DWORD in, out, end, len;

  for (in = out = 0; data[in]; in += end + 1) {
    for (end = 0; data[in+end] && data[in+end] != '.'; end++);
    if (end > 4 && isxnhdr(&data[in])) {
      len = sizeof(dbuf) >> 1;

      if (punycode_decode(end - 4, &data[in+4], &len, dbuf) ||
       out >= 1023) {
        out = 0;
        break;
      }

#ifndef UNICODE
      out += WideCharToMultiByte(CP_ACP, 0, dbuf, len, data + out,
       1023 - out, NULL, NULL);
#else
      if (len > 1023 - out) len = 1023 - out;
      CopyMemory(data + out, dbuf, len * sizeof(TCHAR));
      out += len;
#endif
    }
    else {
      if (in != out) MoveMemory(data + out, data + in, end * sizeof(TCHAR));
      out += end;
    }

    if (!data[in+end]) break;
    data[out++] = data[in+end];
  }

  data[out] = 0;
  return 3;
}

#ifdef UNICODE
typedef struct {
  DWORD  mVersion;
  HWND   mHwnd;
  BOOL   mKeep;
  BOOL   mUnicode;
} LOADINFO;

void __export __stdcall LoadDll(LOADINFO *li) {
  li->mKeep = FALSE;
  li->mUnicode = TRUE;
}
#endif
